#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Sérialisation
# 
# Le notebook explore différentes façons de sérialiser des données et leurs limites.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## JSON
# 
# Le format [JSON](https://fr.wikipedia.org/wiki/JavaScript_Object_Notation) est le format le plus utilisé sur internet notemmant via les [API REST](https://fr.wikipedia.org/wiki/Representational_state_transfer).

# ### Ecriture

# In[2]:


data = {'records': [{'nom': 'Xavier', 'prénom': 'Xavier', 
                     'langages':[{'nom':'C++', 'age':40}, {'nom':'Python', 'age': 20}]}]}


# In[3]:


from json import dump
from io import StringIO
buffer = StringIO()
res = dump(data, buffer)  # 1
seq = buffer.getvalue()
seq


# ### Lecture

# In[4]:


from json import load
buffer = StringIO(seq)
read = load(buffer)
read


# ### Limite
# 
# Les matrices [numpy](http://www.numpy.org/) ne sont pas sérialisables facilement.

# In[5]:


import numpy
data = {'mat': numpy.array([0, 1])}

buffer = StringIO()
try:
    dump(data, buffer)
except Exception as e:
    print(e)


# Les classes ne sont pas sérialisables non plus facilement.m

# In[6]:


class A:
    def __init__(self, att):
        self.att = att
        
data = A('e')
buffer = StringIO()
try:
    dump(data, buffer)
except Exception as e:
    print(e)        


# Pour ce faire, il faut indiquer au module [json](https://docs.python.org/3/library/json.html) comment convertir la classe en un ensemble de listes et dictionnaires et la classe [JSONEncoder](https://docs.python.org/3/library/json.html#json.JSONEncoder).

# In[7]:


from json import JSONEncoder
class MyEncoder(JSONEncoder):
        def default(self, o):
            return {'classname': o.__class__.__name__, 'data': o.__dict__}
        
data = A('e')
buffer = StringIO()
res = dump(data, buffer, cls=MyEncoder)
res = buffer.getvalue()
res


# Et la relecture avec la classe [JSONDecoder](https://docs.python.org/3/library/json.html#json.JSONDecoder).

# In[8]:


from json import JSONDecoder

class MyDecoder(JSONDecoder):
        def decode(self, o):
            dec = JSONDecoder.decode(self, o)
            if isinstance(dec, dict) and dec.get('classname') == 'A':
                return A(dec['data']['att'])
            else:
                return dec
        
buffer = StringIO(res)
obj = load(buffer, cls=MyDecoder)
obj


# ### Sérialisation rapide
# 
# Le module [json](https://docs.python.org/3/library/json.html) est la librairie standard de Python mais comme la sérialisation au format *JSON* est un besoin très fréquent, il existe des alternative plus rapide comme [ujson](https://docs.micropython.org/en/latest/pyboard/library/ujson.html).

# In[9]:


data = {'records': [{'nom': 'Xavier', 'prénom': 'Xavier', 
                     'langages':[{'nom':'C++', 'age':40}, {'nom':'Python', 'age': 20}]}]}


# In[10]:


get_ipython().run_line_magic('timeit', 'dump(data, StringIO())')


# In[11]:


from ujson import dump as udump
get_ipython().run_line_magic('timeit', 'udump(data, StringIO())')


# Ces deux lignes mesures l'écriture au format JSON mais il faut aussi mesurer la lecture.

# In[12]:


buffer = StringIO()
dump(data, buffer)
res = buffer.getvalue()
get_ipython().run_line_magic('timeit', 'load(StringIO(res))')


# In[13]:


from ujson import load as uload
get_ipython().run_line_magic('timeit', 'uload(StringIO(res))')


# On enlève le temps passé dans la creation du buffer.

# In[14]:


get_ipython().run_line_magic('timeit', 'StringIO(res)')


# ## Pickle
# 
# Le module [pickle](https://docs.python.org/3/library/pickle.html) effectue la même chose mais au format binaire. Celui-ci est propre à *Python* et ne peut être lu d'autres langages, voire parfois par d'autres versions de .Python*.

# ### Ecriture

# In[15]:


data = {'records': [{'nom': 'Xavier', 'prénom': 'Xavier', 
                     'langages':[{'nom':'C++', 'age':40}, {'nom':'Python', 'age': 20}]}]}


# In[16]:


from pickle import dump
from io import BytesIO
buffer = BytesIO()
res = dump(data, buffer)
seq = buffer.getvalue()
seq


# ### Lecture

# In[17]:


from pickle import load
buffer = BytesIO(seq)
read = load(buffer)
read


# ### Les classes
# 
# A l'inverse du format *JSON*, les classes sont sérialisables avec *pickle* parce que le langage utilise un format très proche de ce qu'il a en mémoire. Il n'a pas besoin de conversion supplémentaire.

# In[18]:


data = A('r')
buffer = BytesIO()
res = dump(data, buffer)
seq = buffer.getvalue()
seq


# In[19]:


buffer = BytesIO(seq)
read = load(buffer)
read


# ### Réduire la taille
# 
# Certaines informations sont duppliquées et il est préférable de ne pas les sérialiser deux fois surtout si elles sont voluminueuses.

# In[20]:


class B:
    def __init__(self, att):
        self.att1 = att
        self.att2 = att


# In[21]:


data = B('r')
buffer = BytesIO()
res = dump(data, buffer)
seq = buffer.getvalue()
seq


# Evitons maintenant de stocker deux fois le même attribut.

# In[22]:


class B:
    def __init__(self, att):
        self.att1 = att
        self.att2 = att
        
    def __getstate__(self):
        return dict(att=self.att1)
    
data = B('r')
buffer = BytesIO()
res = dump(data, buffer)
seq = buffer.getvalue()
seq


# C'est plus court mais il faut inclure maintenant la relecture.

# In[23]:


class B:
    def __init__(self, att):
        self.att1 = att
        self.att2 = att
        
    def __getstate__(self):
        return dict(att=self.att1)
    
    def __setstate__(self, state):
        setattr(self, 'att1', state["att"])
        setattr(self, 'att2', state["att"])
    
buffer = BytesIO(seq)
read = load(buffer)
read


# In[24]:


read.att1, read.att2


# In[25]:


data = B('r')
get_ipython().run_line_magic('timeit', 'dump(data, BytesIO())')


# In[26]:


get_ipython().run_line_magic('timeit', 'load(BytesIO(seq))')


# La sérialisation binaire est habituellement plus rapide dans les langages bas niveau comme C++. La même comparaison pour un langage haut niveau tel que Python n'est pas toujours prévisible. Il est possible d'accélérer un peu les choses.

# In[27]:


from pickle import HIGHEST_PROTOCOL
get_ipython().run_line_magic('timeit', 'dump(data, BytesIO(), protocol=HIGHEST_PROTOCOL)')


# ## Cas des fonctions
# 
# La sérialisation s'applique à des données et non à du code mais le fait de sérialiser des fonctions est tout de même tentant. La sérialisation binaire fonctionne même avec les fonctions.

# ### Binaire

# In[28]:


def myfunc(x):
    return x + 1

data = {'x': 5, 'f': myfunc}

from pickle import dump
from io import BytesIO
buffer = BytesIO()
res = dump(data, buffer)
buffer.getvalue()


# In[29]:


from pickle import load
res = load(BytesIO(buffer.getvalue()))
res


# In[30]:


res['f'](res['x'])


# La sérialisation ne conserve pas le code de la fonction, juste son nom. Cela veut dire que si elle n'est pas disponible lorsqu'elle est appelé, il sera impossible de s'en servir.

# In[31]:


del myfunc

from pickle import load
try:
    load(BytesIO(buffer.getvalue()))
except Exception as e:
    print(e)


# Il est possible de contourner l'obstacle en utilisant le module [cloudpicke](https://github.com/cloudpipe/cloudpickle) qui stocke le code de la fonction.

# In[32]:


def myfunc(x):
    return x + 1

data = {'x': 5, 'f': myfunc}

from cloudpickle import dump
from io import BytesIO
buffer = BytesIO()
res = dump(data, buffer)
buffer.getvalue()


# In[33]:


del myfunc

from cloudpickle import load
res = load(BytesIO(buffer.getvalue()))
res


# In[34]:


res['f'](res['x'])


# ### JSON

# La sérialisation au format JSON ne fonctionne pas avec le module standard.

# In[35]:


from json import dump
from io import StringIO
buffer = StringIO()
try:
    dump(data, buffer)  # 2
except Exception as e:
    print(e)


# La sérialisation avec *ujson* ne fonctionne pas non plus même si elle ne produit pas toujours d'erreur.

# In[36]:


from ujson import dump
from io import StringIO
buffer = StringIO()
try:
    res = dump(data, buffer)  # 3
except TypeError as e:
    print(e)
buffer.getvalue()


# ## Cas des itérateurs
# 
# Les itérateurs fonctionnent avec la sérialisation binaire mais ceci implique de stocker l'ensemble que l'itérateur parcourt.

# In[37]:


ens = [1, 2]

data = {'x': 5, 'it': iter(ens)}

from pickle import dump
from io import BytesIO
buffer = BytesIO()
res = dump(data, buffer)  # 4
buffer.getvalue()


# In[38]:


del ens
from pickle import load
res = load(BytesIO(buffer.getvalue()))
res


# In[39]:


list(res["it"])


# In[40]:


list(res["it"])


# ## Cas des générateurs
# 
# Ils ne peuvent être sérialisés car le langage n'a pas accès à l'ensemble des éléments que le générateur parcourt. Il n'y a aucun moyen de sérialiser un générateur mais on peut sérialiser la fonction qui crée le générateur.

# In[41]:


def ensgen():
    yield 1
    yield 2

data = {'x': 5, 'it': ensgen()}

from pickle import dump
from io import BytesIO
buffer = BytesIO()
try:
    dump(data, buffer)
except Exception as e:
    print(e)


# In[42]: